#Usage: vmd -dispdev text -e $argv0 -args inPrefix outPrefix numMGH

#Adds a user defined amount of magnesium hexahydride to in a user defined box around an input system.
#Requires MGH coordinate and bond files, as well as a topology file. The MGH coordinate file must be in first octant.
#Created by Derek Van Damme on March 13, 2014.
#Updated to version 2 to by Derek Van Damme on April 25, 2014.
#Updated to version 2.2 by Derek Van Damme on May 2,2014.
#Updated to version 2.3 by Derek Van Damme on August 3rd, 2014.

# Modified by Chen-Yu Li cli56@illinois.edu on 2015/5/18


#!Warning! with current method, large temporary files could be created for systems a desiring large amount of MGH groups to be added. (Large being > ~100,000 as an estimate).

set timestart [clock seconds]

package require psfgen
package require solvate
resetpsf

puts "STATUS) Loading input parameters and performing initial calculations."

# ---------------------------------------------------------------------------------------- #
# ------------------ START of INPUT Parameter Section ------------------------------------ #
# ---------------------------------------------------------------------------------------- #

#Define Input files
set inputpsf	[lindex $argv 0].psf
set inputpdb	[lindex $argv 0].pdb

#Define Output name
set outputname [lindex $argv 1]


#Change keepCoord to 1 if an output file containing the accepted coordinates is desired.
set keepCoord 0
if {$keepCoord} {
	#Defines name of the output coordinate file.
	set coordFile acc_coord_${outputname}.log
}

#Define desired number of MGH groups.
#set numMGH [expr total + number + mgh]
set numMGH [lindex $argv 2]


set id [mol load psf $inputpsf pdb $inputpdb]
set all [atomselect top all]
set minmax [measure minmax $all]
mol delete top
set x_min [expr [lindex $minmax 0 0] - 30]
set y_min [expr [lindex $minmax 0 1] - 30]
set z_min [expr [lindex $minmax 0 2] - 30]
set x_max [expr [lindex $minmax 1 0] + 30]
set y_max [expr [lindex $minmax 1 1] + 30]
set z_max [expr [lindex $minmax 1 2] + 30]


#Define Box Bounds. Must be large enough that the input is not within closeInt grids from the boundary.
	#Lower bound
set low_boxBound "$x_min $y_min $z_min"
	#Upper bound
set up_boxBound "$x_max $y_max $z_max"
	#Size of box
set boxSize [vecsub $up_boxBound $low_boxBound]

#Set PDB and PSF defaults.
#NOTE: These files could be easily changed to accommodate adding a different type of molecule to the system. If so desired, remember to change the mghSize parameter appropriately.
	#directory of mgh files.
set mghdir "ff_jejoong"		
	#Default files
set mghpdb "mghh.pdb"		;# Debug 
set mghpsf "mghh.psf"
set mghtop "$mghdir/toppar_water_ions_jejoong.str"

#Boundary area radius for an MGH group measuring from the center (assume spherical). 
set mghSize 4.0

#Safeguard parameter that prevents script from iterating indefinitely.
set itLimit 50000

#If two additional output files of MGH group only pdb and psf formats, set keepMGH to 1. If not, set to 0.
set keepMGH 0

#Define the size of the grid used to find coordinate to place MGH. Needs to be at least mghSize.
set gridSize [expr 1.25 * $mghSize]

#Define how my grids away from input system is considered close.
set closeInt 1

#Specify whether grid lists are desired to be printed to files.
set keepGridList 0

#Specify whether to create empty grid list or not. (necessary for placing MGH).
set keepEmptyList 0

#Specify probability that grids for adding MGH should be chosen exclusively from the close grid list, rather than from all possible empty grids.
#This probability should be around (Neutralizing MGH)/(total MGH).
#set prob [expr NeutralizingMGH/ TotalMGH]
set prob 1 

# ---------------------------------------------------------------------------------------- #
# ---------------------- END of INPUT Parameter Section ---------------------------------- #
# ---------------------------------------------------------------------------------------- #

topology $mghtop

#load system
mol load psf $inputpsf
mol addfile $inputpdb

set system [atomselect top all]

#variables used Loop
set n 0
set g 0
#set coordList { {0 0 0} }
set sumtest1 0
set test1 0
set gridCoord { 0 0 0 }
#set gridAcc 0
#set gridRem 0
#set sumtest2 0
#set test2 0

#Measure time to determine run time of loops. Debug
set timestart [clock seconds]

#Final atoms resid. (Assumes this resid is the largest resid number). Debug
set sysSize [$system num]
incr sysSize -1
set latom [atomselect top "index $sysSize"]
set lastSysRes [$latom get resid]
set mghRes [expr $lastSysRes + 1]
#set mghRes 38	;# placed because 37 in hexadecimal is 25 which would overlap with this particular system.

puts "STATUS) Finished loading input parameters and initial calculations."


# Vol map start
# ------------------------------------------------------------------------------------------------------ #
# --------------- Read coordinates of the input system and specify grid number that they lie in -------- #
# ------------------------------------------------------------------------------------------------------ #

# Break down defined boxsize of system into a grid.

#Determine number of grids in x, y, and z directions.
set numGrid [vecscale [expr 1.0/ $gridSize ] $boxSize]
foreach p {0 1 2} {
	set numGrid [lreplace $numGrid $p $p [expr round( floor( [lindex $numGrid $p] ) ) ] ]
}
#To avoid overextending boundaries.
set numGrid [ vecsub $numGrid {1 1 1} ]
#To convert to integer form.
set numGrid [list [expr round( [lindex $numGrid 0] ) ] [expr round( [lindex $numGrid 1] ) ] [expr round( [lindex $numGrid 2] ) ] ]

#Determine total number of grids.
set numGridTotal [expr [lindex $numGrid 0] * [lindex $numGrid 1] * [lindex $numGrid 2] ]


# Debug
#puts  "Pizza"

#Cycle through coordinates and determine corresponding grid.
foreach coord1 [$system get {x y z} ] {
	
	#Determine difference between coordinate and minimum coordinate of the boundary.
	set coordDiff [vecsub $coord1 $low_boxBound ]

	#Determine each x, y, z grid number.
	foreach xyz {0 1 2} {
		#Determine how many grids the coordinate is away from the lower boundary in current axis.
		set f [ expr [lindex $coordDiff $xyz] / $gridSize ] 

			set gridCoord [lreplace $gridCoord $xyz $xyz [ expr round( floor($f + 1) ) ] ]
		
	}
	
	#Once the x, y, and z grid numbers are determined, determine the overall grid number.
	set gridNum [expr round( [lindex $gridCoord 0] + ( [lindex $numGrid 0] * [lindex $gridCoord 1] ) + ( [lindex $numGrid 0] * [lindex $numGrid 1] * [lindex $gridCoord 2] ) ) ]
	
	#Add grid number to the list of grids to remove.
	lappend gridRem $gridNum
	
}

#Delete system molecules.
mol delete top

unset gridNum gridCoord coord1

# ----------- Reset number of grids in system, ignoring grids that contain system coordinates. -------- #

#sort the list of grids to remove, from lowest integer to highest.
set gridRem [lsort -integer -unique $gridRem]

# Debug
puts "STATUS) Found grids containing input system."

#Cycle through each filled grid and create list of grids that are close (as defined by closeInt) to the input system.
foreach badGrid $gridRem {
	#Determine x, y, and z grid numbers of current filled grid.
	set badGridNum [list [expr ($badGrid%( [lindex $numGrid 0]*[lindex $numGrid 1] ) )%[lindex $numGrid 0] ] [expr ($badGrid%( [lindex $numGrid 0]*[lindex $numGrid 1] ) )/[lindex $numGrid 0] ] [expr $badGrid/( [lindex $numGrid 0]*[lindex $numGrid 1] ) ] ]
		
	#Cycle through all possible grids within closeInt of the current filled grid, and add them to the gridClose list. 
	for {set i 0} {$i <= $closeInt } {incr i} {
		if {$i > $closeInt} {
			set i $closeInt
			}
		for {set j 0} {$j <= $closeInt } {incr j} {
			if {$j > $closeInt} {
				set j $closeInt
			}
			for {set k 0} {$k <= $closeInt } {incr k} {
				if {$k > $closeInt} {
					set k $closeInt
				}
				#set the x, y, and z grid numbers for current permutation.
				set gridCloseNum [list [expr [lindex $badGridNum 0] + $i] [expr [lindex $badGridNum 1] + $j] [expr [lindex $badGridNum 2] + $k] ]
				#Convert x, y, and z grid number into total grid number, and add it to the close grid list.
				lappend gridClose [expr round( [lindex $gridCloseNum 0] + ( [lindex $numGrid 0 ] * [lindex $gridCloseNum 1] ) + ( [lindex $numGrid 0 ] * [lindex $numGrid 1 ] * [lindex $gridCloseNum 2] ) ) ]
			}
		}
	}
	#Cycle through all possible grids within closeInt of the current filled grid, and add them to the gridClose list. 
	for {set i [expr -1 * $closeInt] } {$i <= 0 } {incr i} {
		if { [lindex $badGridNum 0] + $i <= 0} {
		} else {
			for {set j [expr -1 * $closeInt]} {$j <= 0 } {incr j} {
				if { [lindex $badGridNum 1] + $j <= 0} {
				} else {
					for {set k [expr -1 * $closeInt] } {$k <= 0 } {incr k} {
						if { [lindex $badGridNum 2] + $k <= 0} {
						} else {
							#Skip coordinate if it is outside the original user defined boundaries. Debug
							
							#set the x, y, and z grid numbers for current permutation.
							set gridCloseNum [list [expr [lindex $badGridNum 0] + $i] [expr [lindex $badGridNum 1] + $j] [expr [lindex $badGridNum 2] + $k] ]
							#Convert x, y, and z grid number into total grid number, and add it to the close grid list.
							lappend gridClose [expr round( [lindex $gridCloseNum 0] + ( [lindex $numGrid 0 ] * [lindex $gridCloseNum 1] ) + ( [lindex $numGrid 0 ] * [lindex $numGrid 1 ] * [lindex $gridCloseNum 2] ) ) ]
						}
					}
				}
			}
		}
	}
}


unset i j k badGrid badGridNum gridCloseNum

set gridClose [lsort -integer -unique $gridClose]

set i 0
#Remove filled grids from close grid.
foreach gridC $gridClose {
	foreach badGrid $gridRem {
		if {$badGrid == $gridC} {
			set gridClose [lreplace $gridClose $i $i -2]
			break
		}
	}
	
	incr i
}
set gridClose [lsort -integer -unique $gridClose]
set gridClose [lreplace $gridClose 0 0]

unset i gridC badGrid

# Debug
puts "STATUS) Determined grids within $closeInt grid(s) of grids containing input."

if {$keepEmptyList == 1} {

	#Initialize list of grids not containing input system, by listing all grids at first.
	#After cycling through input coordinates, grids containing the input system will be removed.
	for {set i 1} { $i <= $numGridTotal } {incr i} {
		lappend gridAcc $i
	}

	#Remove filled grids, and close grids from gridAcc, by replacing them with the value of -2.
	#Then unique integer sort the acceptable list and remove the first entry (which will be -2).
	#The index of the unacceptable grids in gridAcc should be 1 less than the value of the grid number.
	foreach badGrid $gridRem {
		if { $badGrid <= $numGridTotal} {
			set gridAcc [lreplace $gridAcc [expr $badGrid - 1]  [expr $badGrid - 1] -2]
		}
	}
	foreach gridC $gridClose {
		if {$gridC <= $numGridTotal} {
			set gridAcc [lreplace $gridAcc [expr $gridC - 1]  [expr $gridC - 1] -2]
		}
	}
	
	set gridAcc [lsort -integer -unique $gridAcc]
	set gridAcc [lreplace $gridAcc 0 0]

	unset badGrid gridC

	# Debug
	puts "STATUS) Determined empty grids that are not close to input system."
	
}


#Print grid lists to file if specified in input section.
if {$keepGridList == 1} {

	set fileID1 [open ${outputname}_filledGridsi.dat "w"]
	foreach line $gridRem {
		puts $fileID1 "$line "
	}
	close $fileID1
	
	if {$keepEmptyList == 1} {
		set fileID2 [open ${outputname}_emptyGrids.dat "w"]
		foreach line $gridAcc {
			puts $fileID2 "$line "
		}
		close $fileID2

	}
	
	set fileID3 [open ${outputname}_closeGridsi.dat "w"]
	foreach line $gridClose {
		puts $fileID3 "$line "
	}
	close $fileID3	
	
	unset fileID1 fileID2 fileID3 line
	puts "STATUS) Grid number lists written to file."
}

puts "STATUS) Finished with vol_map."


# vol map end
# ---------------------------------------------------------------------------------------------------------------
# ------------ Randomly choose grids to place MGH into. Create preference for grids closer to input system. Ignore previously used grids. ---- #
# ----------------------------------------------------------------------------------------------------------------


puts "STATUS) Determining coordinates to add MGH."

#Determine number of acceptable grids.
set numAcc [llength $gridAcc]

# End program if not enough grids to place number of desired MGH. Debug
if { $numMGH >= $numAcc} {
	puts "ERROR: Not enough grids accepted to place $numMGH MGH. Please decrease numMGH input or decrease gridSize input."
	exit
}

# gridClose and gridAcc will be have values removed in this stage. Debug

#Generate MGH coordinates until user defined number desired (numMGH) is reached or if maximum allowed iterations limit (itLimit) is reached.
for {set i 0} {$i < $numMGH && $i <= $itLimit } {incr i} {
	
	#Randomly decide whether to choose grid from close grid list or empty grid list, with distribution specified by user input prob.
	if { rand() <= $prob } {
		set range [llength $gridClose]
		#Randomly choose index number to select from close grid list.
		set ind [expr int( ($range+0.999) * rand() ) ]
		#Add grid to list.
		set gridAdd [lindex $gridClose $ind ]
		#Remove chosen grid from list to avoid repeat selections.
		set gridClose [lreplace $gridClose $ind $ind]
		
		
	} else {
		set range [llength $gridAcc]
		#Randomly choose index number to select from empty grid list.
		set ind [expr int( ($range+1) * rand() ) ]
		#Add grid to list.
		set gridAdd [lindex $gridAcc $ind ]
		#Remove chosen grid from list to avoid repeat selections.
		set gridAcc [lreplace $gridAcc $ind $ind]
	}

	
	#Determine x, y, and z grid numbers.
	set gridAddNum [list [expr ($gridAdd%( [lindex $numGrid 0]*[lindex $numGrid 1] ) )%[lindex $numGrid 0] ] [expr ($gridAdd%( [lindex $numGrid 0]*[lindex $numGrid 1] ) )/[lindex $numGrid 0] ] [expr $gridAdd/( [lindex $numGrid 0]*[lindex $numGrid 1] ) ] ]
	
	#Create coordinates in center of the accepted grid.
	set coordAdd [vecadd $low_boxBound [vecsub [vecscale $gridSize $gridAddNum] [list [expr (0.5*$mghSize)+(($gridSize - $mghSize)*rand() ) ] [expr (0.5*$mghSize)+(($gridSize - $mghSize)*rand() ) ] [expr (0.5*$mghSize)+(($gridSize - $mghSize)*rand() ) ] ] ] ]
	
	#Add coordinates to generated accepted coordinate list.
	lappend coordList $coordAdd
}


# ------ Debug
#Opens file to write the accepted coordinates in.
if {$keepCoord} {
	set fileID4 [open $coordFile "w"]

	foreach line $coordList {

		puts $fileID4 "$line "
	}

	close $fileID4
}
# ------

puts "STATUS) Coordinates found."

# ---------------------------------------------------------------------------------------- #
# -------------------------- Creating PDB and PSF Files ---------------------------------- #
# ---------------------------------------------------------------------------------------- #


# Delete variables to free up memory. Debug
unset range fileID4 line gridAddNum coordAdd ind numAcc


#Display an error and exit if script wasn't finished because the iteration limit was reached.
if {$i >= $itLimit} {
	puts "ERROR: Iteration limit reached"
	
	exit
} else {
	
	puts "STATUS) Creating PDB and PSF files."
	
	#index variable
	set j 0
	set w [expr $j + 1]		;# Debug

	#Create acceptable MGH segment by creating an MGH group at each coordinate and adding them to the segment structure, iteratively.
	segment MGHH {	
		#Create MGH groups at accepted coordinates.
		foreach accCoord $coordList {
			
			#resetpsf	;# Debug
			
			#Load coordinates.
			mol load pdb $mghpdb
				
			#select MGH group and move it to current generated coordinates.		Debug is moving these coordinates redundant if they don't get loaded in this time?
			set mghMol [atomselect top all]
			$mghMol moveby [lindex $coordList $j]
				
				
			#Change Resid (in hexadecimal format) of MGH group. Debug
			#set hex [format %X $mghRes]
			#$mghMol set resid $hex	;# For hexadecimal residue ID
			$mghMol set resid $mghRes	;# For integer residue ID
				
			#write the updated MGH group with new coordinates to file.
			$mghMol writepdb mgh_one_tmp.pdb
			
			#Delete MGH group
			mol delete top
			
			#load current mgh group structure to the segment
			pdb mgh_one_tmp.pdb
				
				
			#increase index
			incr j

			#Increase the resid number to use for next MGH group. Debug
			incr mghRes
				
		}
	}

	#index variable
	set j 0
	set w [expr $j + 1]		;# Debug
	set mghRes [expr $lastSysRes + 1]
	# ### Debug
	foreach accCoord $coordList {
			
			#resetpsf	;# Debug
			
			#Load coordinates.
			mol load pdb $mghpdb
				
			#select MGH group and move it to current generated coordinates.
			set mghMol [atomselect top all]
			$mghMol moveby [lindex $coordList $j]
				
				
			#Change Resid (in hexadecimal format) of MGH group. Debug
			#set hex [format %X $mghRes]
			#$mghMol set resid $hex	;# For hexadecimal residue ID
			$mghMol set resid $mghRes	;# For integer residue ID
				
			#write the updated MGH group with new coordinates to file.
			$mghMol writepdb mgh_one_tmp.pdb
			
			#Delete MGH group
			mol delete top
				
			#Adds coordinates to the structure.	
			coordpdb mgh_one_tmp.pdb	;# Debug	
				
			#increase index
			incr j

			#Increase the resid number to use for next MGH group. Debug
			incr mghRes
				
	}
	# #####		
}

					
		
	#writepsf ${outputname}_tmp.psf
	#writepdb ${outputname}_tmp.pdb
	writepsf mgh_group_tmp.psf
	writepdb mgh_group_tmp.pdb
	# #####

	#write output files
	resetpsf
	readpsf $inputpsf
	coordpdb $inputpdb
	readpsf mgh_group_tmp.psf	;# Debug
		

	coordpdb mgh_group_tmp.pdb
		
	writepsf ${outputname}.psf
	writepdb ${outputname}.pdb
		
	#Choose whether to keep MGH only pdb and psf files
	if {$keepMGH} {
		
		file copy -force mgh_group_tmp.psf mgh_only_${outputname}.psf
		file copy -force mgh_group_tmp.pdb mgh_only_${outputname}.pdb
			
	}

#Delete temporary files
file delete mgh_group_tmp.psf
file delete mgh_group_tmp.pdb
file delete mgh_one_tmp.pdb


puts "STATUS) Finished creating PDB and PSF coordinates."

	
#Time for script. Debug
set timeend [clock seconds]
#puts "Run time: "
set runtime [expr $timeend - $timestart]


puts "STATUS) Process Completed."
puts "STATUS) Total runtime: $runtime seconds"

exit
